"""Console script for spyci."""
import argparse
import textwrap
import sys

from spyci import spyci


def parse_args():
    description = textwrap.dedent("""
        Spyci (spyci v0.6.1) -- parses ngspice raw data files and
        plots the specified variables.
        For full documentation check the repo: https://github.com/gmagno/spyci
    """)

    epilog = r"""
                                                 /##
                                                |__/
          /#######  /######  /##   /##  /####### /##
         /##_____/ /##__  ##| ##  | ## /##_____/| ##
        |  ###### | ##  \ ##| ##  | ##| ##      | ##
         \____  ##| ##  | ##| ##  | ##| ##      | ##
         /#######/| #######/|  #######|  #######| ##
        |_______/ | ##____/  \____  ## \_______/|__/
                  | ##       /##  | ##              
                  | ##      |  ######/              
                  |__/       \______/               
    """

    epilog += textwrap.dedent("""
        return:
            The return value of %(prog)s is 0 if the raw file is successfully
            parsed and plotted.

        examples:
            # Run without arguments will attempt to load rawspice.raw from cwd
            # and plot all variables
            $ %(prog)s

            # List variables that can be plotted
            $ %(prog)s -l
            Variables:

            idx  name        type
            -----  ----------  -------
                1  i(l1)       current
                2  n1          voltage
                3  vi          voltage
                4  vo          voltage
                5  i(vsource)  current

            # Load 'some/location/sim.raw' and plot variables 'i(l1)' and 'vo'
            $ %(prog)s -r some/location/sim.raw "i(l1)" vo

            # Indices can be used insted of variable names, this is equivalent
            # to the previous example
            $ %(prog)s -r some/location/sim.raw 1 4

            # Save your plot to the file system
            $ %(prog)s -o myplot.png 1 4

            # Different image formats are supported, just use the correct
            # extension, {.png, .svg, .pdf, ...}. For a list of supported
            # formats run with -f flag
            $ %(prog)s -f
            Supported output image file formats:

            ext    format
            -----  -------------------------
            raw    Raw RGBA bitmap
            rgba   Raw RGBA bitmap
            pgf    PGF code for LaTeX
            svgz   Scalable Vector Graphics
            svg    Scalable Vector Graphics
            ps     Postscript
            png    Portable Network Graphics
            eps    Encapsulated Postscript
            pdf    Portable Document Format

        copyright:
            Copyright © 2020 Gonçalo Magno <goncalo@gmagno.dev>
            This software is licensed under the MIT License.
    """)
    parser = argparse.ArgumentParser(
        prog="spyci",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=description,
        epilog=epilog)
    parser.add_argument("-v",
                        "--version",
                        required=False,
                        action='store_true',
                        help="shows %(prog)s version")
    parser.add_argument("-r",
                        "--raw-file",
                        required=False,
                        default="rawspice.raw",
                        action='store',
                        help="path to raw file to be parsed")
    parser.add_argument("-l",
                        "--list-variables",
                        required=False,
                        action='store_true',
                        help="lists variables that can be plotted")
    parser.add_argument("-f",
                        "--out-formats",
                        required=False,
                        action='store_true',
                        help="lists supported output image formats")
    parser.add_argument("-o",
                        "--out-image",
                        required=False,
                        action='store',
                        help="path to output image file, use -f, to list "
                        "supported formats")
    parser.add_argument("vars",
                        metavar='VARS',
                        nargs=argparse.REMAINDER,
                        help="List of variables to plot")
    args = vars(parser.parse_args())
    return args


def main():
    kwargs = parse_args()
    if kwargs['version']:
        print("Spice Raw Parse -- spyci v0.6.1")
        return 0

    if kwargs['list_variables']:
        try:
            spyci.list_vars(kwargs['raw_file'])
        except FileNotFoundError as e:
            print("Error: {}: {}".format(e.strerror, e.filename))
            return -1
        return 0

    if kwargs['out_formats']:
        try:
            spyci.img_formats()
        except FileNotFoundError as e:
            print("Error: {}: {}".format(e.strerror, e.filename))
            return -1
        return 0

    try:
        spyci.plot(kwargs['raw_file'], kwargs['vars'], kwargs['out_image'])
    except FileNotFoundError as e:
        print("Error: {}: {}".format(e.strerror, e.filename))
        return -1
    except spyci.InvalidVarsError:
        print("Error: Could not find variables: {}. For a list of valid "
              "variables run with '-l' flag".format(kwargs['vars']))
        return -1

    return 0


if __name__ == "__main__":
    sys.exit(main())  # pragma: no cover



"""Main module."""

import copy
import os
import pathlib
import re

import matplotlib.pyplot as plt
import numpy as np
from tabulate import tabulate


class InvalidVarsError(Exception):
    pass


def load_raw(rawfile):
    """
    Parses an ascii raw data file and returns a dictionary with the following
    structure:
        {
            'title': <str>,
            'date:': <str>,
            'plotname:': <str>,
            'flags:': <str>,
            'no_vars:': <str>,
            'no_points:': <str>,
            'vars': [
                { 'idx': <int>, 'name': <str>, 'type': <str> },
                { 'idx': <int>, 'name': <str>, 'type': <str> }
                ...
                { 'idx': <int>, 'name': <str>, 'type': <str> }
            ]
            'values': {
                'var1': <numpy.ndarray>,
                'var2': <numpy.ndarray>,
                ...
                'varN': <numpy.ndarray>
            }
        }

    Arguments:
        :filename: path to file with raw data.
    Returns
        dict with structure described above.
    """

    path = os.path.realpath(rawfile)
    data = pathlib.Path(path).read_text()
    ret = {}
    pattern = (r"^Title:\s*(?P<title>.*)\s*"
               r"^Date:\s*(?P<date>\w.*?\w)\s*"
               r"Plotname:\s*(?P<plotname>\w.*?\w)\s*"
               r"Flags:\s*(?P<flags>\w.*?\w)\s*"
               r"No\. Variables:\s*(?P<no_vars>\d+)\s*"
               r"No\. Points:\s*(?P<no_points>\d+)\s*"
               r"Variables:\s*$(?P<vars>.*)^"
               r"Values:(?P<values>.*)")
    m = re.search(pattern, data, re.MULTILINE | re.DOTALL)
    ret = copy.deepcopy(m.groupdict())
    ret.pop('values')

    # vars
    vars = m.groupdict()['vars']
    pattern = (r"\s*(?P<idx>\d+)\s+" r"(?P<name>\S+)\s+" r"(?P<type>.*)\s*")
    m_vars = re.finditer(pattern, vars)
    ret['vars'] = []
    for i in m_vars:
        ret['vars'].append(i.groupdict())

    # values
    values = m.groupdict()['values']
    pattern = (r"^\s*(?P<idx>\d+)\s+" r"(?P<values>\S+(?:\n\s*\S+)*)$")
    m_values = re.finditer(pattern, values, re.MULTILINE)
    values = []
    for i in m_values:
        e = i.groupdict()['values'].split()
        values.append(tuple(e))
    values = [
        tuple(complex(*map(float, val.split(','))) for val in tup)
        for tup in values
    ]
    dtype = []
    for x in ret['vars']:
        dtype.append((x['name'], np.complex_))
    values = np.array(values, dtype=dtype)
    ret['values'] = values
    return ret


def list_vars(rawfile):
    """Loads the rawspice file and prints the variables to plot.
        Args
        rawfile: path to the rawspice.raw file
    """
    data = load_raw(rawfile)
    pvars = sorted(data['vars'], key=lambda v: v['idx'])
    pvars.pop(0)

    trows = [[v['idx'], v['name'], v['type']] for v in pvars]
    theader = ['idx', 'name', 'type']

    print("Variables:\n")
    print(tabulate(trows, headers=theader))


def img_formats():
    formats = plt.gcf().canvas.get_supported_filetypes()
    trows = [[k, v] for k, v in formats.items()]
    theader = ['ext', 'format']
    print("Supported output image file formats:\n")
    print(tabulate(trows, headers=theader))


def plot(rawfile, pvars=None, outimg=None):
    """Loads the rawspice file and plots the data.
        Args
        rawfile: path to the rawspice.raw file
        pvars: list of variables to plot. If `None`, all variables are
        plotted.
        save: path to save file
    """
    # FIXME: some poop logic here... should definitely fix this
    d = load_raw(rawfile)
    file_vars = sorted(d['vars'], key=lambda v: v['idx'])

    if pvars:
        pvars = [
            v for v in file_vars if v['name'] in pvars or v['idx'] in pvars
        ]
    else:
        pvars = file_vars
        pvars.pop(0)  # pop first variable, either 'time' or 'frequency'

    if not pvars:
        raise InvalidVarsError()

    fig, ax1 = plt.subplots()
    ax2 = ax1.twinx()

    if d['plotname'] == "Transient Analysis":
        xx = d['values']['time']
        ax1.set_xscale('linear')
        ax1.set_yscale('linear')
        ax2.set_yscale('linear')
        ax1.set_xlabel("Time (s)")
    elif d['plotname'] == "AC Analysis":
        xx = d['values']['frequency']
        ax2.set_xscale('log')
        ax1.set_yscale('log')
        ax2.set_yscale('log')
        ax1.set_xlabel("Frequency (Hz)")
    else:
        print("Error: unsupported plotname")
        return -1

    ax1.set_ylabel("Voltage (V)")
    ax2.set_ylabel("Current (A)")

    lines1 = []
    lines2 = []
    for i, v in enumerate(pvars):
        if v['type'] == 'voltage':
            ln1 = ax1.plot(xx.real,
                           d['values'][v['name']].real,
                           label=v['name'],
                           color=plt.cm.tab10(i))
            lines1.append(ln1)
        elif v['type'] == 'current':
            ln2 = ax2.plot(xx.real,
                           d['values'][v['name']].real,
                           label=v['name'],
                           color=plt.cm.tab10(i))
            lines2.append(ln2)

    ax1.tick_params(axis='x', colors='black')
    ax1.xaxis.label.set_color('black')
    if lines1:
        ax1.legend(loc='upper left')
        ax1.grid(True)
        ax1.tick_params(axis='y', colors='blue')
        ax1.yaxis.label.set_color('blue')
    else:
        ax1.tick_params(axis='y', colors='white')
        ax1.yaxis.label.set_color('white')
    if lines2:
        ax2.legend(loc='upper right')
        if not lines1:
            ax1.xaxis.grid(True)
            ax2.yaxis.grid(True)
        ax2.tick_params(axis='y', colors='red')
        ax2.yaxis.label.set_color('red')
    else:
        ax2.tick_params(axis='y', colors='white')
        ax2.yaxis.label.set_color('white')

    plt.title("{}\n{}, {}".format(d['title'], d['plotname'], d['date']),
              color='black')

    if outimg:
        plt.savefig(outimg, bbox_inches='tight')

    fig.tight_layout()
    plt.show()
